@@ -1,14 +1,111 @@
+from collections import defaultdict
 from glob import glob
+import logging
 import os
 import posixpath
 import socket
 import sys
+from typing import Sequence, Dict
+
 from raft import task
 import yaml
 from boto3 import Session
+from sewer.auth import ErrataItemType
 from sewer.client import Client
-from sewer.config import ACME_DIRECTORY_URL_STAGING
-from sewer.dns_providers.route53 import Route53Dns
+from sewer.config import ACME_DIRECTORY_URL_STAGING  # noqa: F401, pylint: disable=unused-import
+from sewer.dns_providers.common import dns_challenge
+from sewer.dns_providers.route53 import Route53Dns as SewerRoute53Dns
+
+
+log = logging.getLogger(__name__)
+
+
+class Route53Dns(SewerRoute53Dns):
+    def __init__(self, profile=None):
+        super().__init__()
+        self.session = Session(profile_name=profile)
+        self.r53 = self.session.client('route53', config=self.aws_config)
+        self.waiter = self.r53.get_waiter('resource_record_sets_changed')
+        self.resource_records = defaultdict(set)
+
+    def setup(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
+        for x in challenges:
+            domain_name = x['ident_value']
+            value = dns_challenge(x['key_auth'])
+            challenge_domain = f'_acme-challenge.{domain_name}.'
+            self.resource_records[challenge_domain].add(value)
+        self.create_dns_record(domain_name=None, domain_dns_value=None)
+        return []
+
+    def change_batch(self, action, changes):
+        return {
+            'Comment': f'letsencrypt dns certificate validation {action}',
+            'Changes': [{
+                'Action': action,
+                'ResourceRecordSet': {
+                    'Name': name,
+                    'Type': 'TXT',
+                    'TTL': self.ttl,
+                    'ResourceRecords': [
+                        dict(Value=f'"{value}"')
+                        for value in values
+                    ],
+                },
+            } for name, values in changes.items()],
+        }
+
+    def by_zone_id(self):
+        by_zone_id = defaultdict(lambda: defaultdict(set))
+        for name, values in self.resource_records.items():
+            zone_id = self._find_zone_id_for_domain(name)
+            by_zone_id[zone_id][name] = values
+        return by_zone_id
+
+    def wait(self, zone_id, change_id):
+        log.info('[route53 / %s] waiting for %s', zone_id, change_id)
+        self.waiter.wait(Id=change_id, WaiterConfig=dict(
+            Delay=5,
+            MaxAttempts=24,
+        ))
+        log.info('[route53 / %s] change is complete', zone_id)
+
+    def create_dns_record(self, domain_name=None, domain_dns_value=None):
+        result = {}
+        if domain_name and domain_dns_value:
+            self.resource_records[domain_name].add(domain_dns_value)
+        for name, values in self.resource_records.items():
+            log.info('[route53] adding TXT %s => %s', name, values)
+        for zone_id, changes in self.by_zone_id().items():
+            response = self.r53.change_resource_record_sets(
+                HostedZoneId=zone_id,
+                ChangeBatch=self.change_batch('UPSERT', changes),
+            )
+            change_id = response['ChangeInfo']['Id']
+            result[zone_id] = change_id
+        for zone_id, change_id in result.items():
+            self.wait(zone_id, change_id)
+        return result
+
+    def clear(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
+        self.delete_dns_record(None, None)
+        return []
+
+    def delete_dns_record(self, domain_name=None, domain_dns_value=None):
+        result = {}
+        if domain_name and domain_dns_value:
+            self.resource_records[domain_name].add(domain_dns_value)
+        for name, values in self.resource_records.items():
+            log.info('[route53] removing TXT %s => %s', name, values)
+        for zone_id, changes in self.by_zone_id().items():
+            response = self.r53.change_resource_record_sets(
+                HostedZoneId=zone_id,
+                ChangeBatch=self.change_batch('DELETE', changes),
+            )
+            change_id = response['ChangeInfo']['Id']
+            result[zone_id] = change_id
+        for zone_id, change_id in result.items():
+            self.wait(zone_id, change_id)
+        return result
 
 
 def new_cert(hostname, alt_domains, email=None, profile=None):
@@ -27,14 +124,12 @@ def new_cert(hostname, alt_domains, email=None, profile=None):
         the name of the aws profile to use to connect boto3 to
         appropriate credentials
     """
-    if profile:
-        os.environ['AWS_PROFILE'] = profile
     alt_domains = alt_domains.split(',') if alt_domains else []
     client = Client(
         hostname, domain_alt_names=alt_domains, contact_email=email,
-        provider=Route53Dns(), ACME_AUTH_STATUS_WAIT_PERIOD=5,
+        provider=Route53Dns(profile), ACME_AUTH_STATUS_WAIT_PERIOD=5,
         ACME_AUTH_STATUS_MAX_CHECKS=180, ACME_REQUEST_TIMEOUT=60,
-        LOG_LEVEL='DEBUG')
+        LOG_LEVEL='INFO')
     certificate = client.cert()
     account_key = client.account_key
     key = client.certificate_key
@@ -50,14 +145,14 @@ def get_certificate(ns, hostname, profile=None):
         ssm = session.client('ssm')
         name = '/'.join([ ns, 'apps_keystore', hostname, 'account_key' ])
         account_key = get_chunked_ssm_parameter(name, profile=profile)
-        print('account key retrieved')
+        log.info('account key retrieved')
         name = '/'.join([ ns, 'apps_keystore', hostname, 'key' ])
         response = ssm.get_parameter(Name=name, WithDecryption=True)
         key = response['Parameter']['Value']
-        print('private key retrieved')
+        log.info('private key retrieved')
         name = '/'.join([ ns, 'apps_keystore', hostname, 'cert' ])
         certificate = get_chunked_ssm_parameter(name, profile=profile)
-        print('public cert retrieved')
+        log.info('public cert retrieved')
     except:  # noqa: E722, pylint: disable=bare-except
         account_key = None
         key = None
@@ -66,8 +161,11 @@ def get_certificate(ns, hostname, profile=None):
 
 
 def get_file_from_s3(s3, bucket, ns, filename, decode=True):
-    key = posixpath.join(ns, filename)
-    print(f'retrieving s3://{bucket}/{key}')
+    filename = filename.replace('*', 'star')
+    key = filename
+    if ns:
+        key = posixpath.join(ns, key)
+    log.info('retrieving s3://%s/%s', bucket, key)
     response = s3.get_object(Bucket=bucket, Key=key)
     data = response['Body'].read()
     if decode:
@@ -84,26 +182,26 @@ def get_certificate_from_s3(bucket, ns, hostname, profile=None):
         session = Session(profile_name=profile)
         s3 = session.client('s3')
     except Exception as ex:  # noqa: E722, pylint: disable=broad-except,
-        print(f'exception connecting to s3: {ex}')
-        return
+        log.info('exception connecting to s3: %s', ex)
+        return certificate, account_key, key_content
 
     try:
-        account_key = get_file_from_s3(s3, bucket, ns, f'{hostname}.account_key')
-        print('account key retrieved')
+        account_key = get_file_from_s3(s3, bucket, None, 'global.account_key')
+        log.info('account key retrieved')
     except Exception as ex:  # noqa: E722, pylint: disable=broad-except,
-        print(f'exception getting account key: {ex}')
+        log.info('exception getting account key: %s', ex)
 
     try:
         key_content = get_file_from_s3(s3, bucket, ns, f'{hostname}.key')
-        print('private key retrieved')
+        log.info('private key retrieved')
     except Exception as ex:  # noqa: E722, pylint: disable=broad-except,
-        print(f'exception getting private key: {ex}')
+        log.info('exception getting private key: %s', ex)
 
     try:
         certificate = get_file_from_s3(s3, bucket, ns, f'{hostname}.crt')
-        print('public cert retrieved')
+        log.info('public cert retrieved')
     except Exception as ex:  # noqa: E722, pylint: disable=broad-except,
-        print(f'exception retrieving public cert: {ex}')
+        log.info('exception retrieving public cert: %s', ex)
     return certificate, account_key, key_content
 
 
@@ -114,7 +212,7 @@ def get_chunked_ssm_parameter(name, profile=None):
     for n in range(1, 10):
         try:
             st = f'{name}{n}'
-            print(f'[ssm]  getting {st}')
+            log.info('[ssm]  getting %s', st)
             response = ssm.get_parameter(Name=st, WithDecryption=True)
             rg.append(response['Parameter']['Value'])
         except:  # noqa: E722, pylint: disable=bare-except
@@ -128,17 +226,13 @@ def get_pfx(ns, hostname, bucket, profile=None):
     s3 = session.client('s3')
     hostname = hostname.replace('*', 'star')
     pfx_data = get_file_from_s3(s3, bucket, ns, f'{hostname}.pfx', False)
-    print(f'[pfx]  read {len(pfx_data)} bytes')
+    log.info('[pfx]  read %s bytes', len(pfx_data))
     return pfx_data
 
 
 def renew_cert(
         ns, hostname, alt_domains=None,
-        email=None, profile=None, **kwargs):
-    if profile:
-        os.environ['AWS_PROFILE'] = profile
-    bucket = kwargs.get('bucket')
-    tmp_dir = kwargs.get('tmp_dir')
+        email=None, bucket=None, tmp_dir=None, profile=None, **kwargs):
     if alt_domains:
         if isinstance(alt_domains, str):
             alt_domains = alt_domains.split(',')
@@ -147,14 +241,13 @@ def renew_cert(
     _, account_key, key = get_certificate_from_s3(bucket, ns, hostname, profile)
     client = Client(
         hostname, domain_alt_names=alt_domains, contact_email=email,
-        provider=Route53Dns(), account_key=account_key,
+        provider=Route53Dns(profile), account_key=account_key,
         certificate_key=key, ACME_AUTH_STATUS_WAIT_PERIOD=5,
-        ACME_AUTH_STATUS_MAX_CHECKS=360, ACME_REQUEST_TIMEOUT=3,
-        LOG_LEVEL='DEBUG')
-    response = client.acme_register()
+        ACME_AUTH_STATUS_MAX_CHECKS=360, ACME_REQUEST_TIMEOUT=3)
     if not account_key:
+        client.acme_register()
         content = client.account_key
-        save_account_key(bucket, ns, hostname, content, tmp_dir, profile)
+        save_account_key(bucket, ns, content, tmp_dir, profile)
     if not key:
         client.create_certificate_key()
         content = client.certificate_key
@@ -165,15 +258,14 @@ def renew_cert(
     return certificate, account_key, key
 
 
-def save_account_key(bucket, ns, hostname, content, tmp_dir, profile):
+def save_account_key(bucket, ns, content, tmp_dir, profile):
     session = Session(profile_name=profile)
     s3 = session.client('s3')
-    hostname = hostname.replace('*', 'star')
-    filename = f'{hostname}.account_key'
+    filename = 'global.account_key'
     save_to_temp(tmp_dir, filename, content)
     if isinstance(content, str):
         content = content.encode('utf-8')
-    s3_key = posixpath.join(ns, filename)
+    s3_key = filename
     s3.put_object(Bucket=bucket, Key=s3_key, Body=content)
 
 
@@ -255,13 +347,18 @@ def renew_all(ctx, dir_name=None, profile=None):
     dir_name = os.path.join(dir_name, '*.yml')
     files = glob(dir_name)
     for filename in files:
-        if filename.endswith('defaults.yml'):
-            continue
-        request_cert(ctx, filename, profile, defaults)
+        try:
+            # don't let the failure of any one certificate
+            # make it so that we don't try to renew the rest
+            if filename.endswith('defaults.yml'):
+                continue
+            request_cert(ctx, filename, profile, defaults)
+        except:
+            pass
 
 
 def request_cert(ctx, filename, profile, defaults):
-    print(f'processing {filename}')
+    log.info('processing %s', filename)
     with open(filename, 'r') as f:
         values = yaml.load(f, Loader=yaml.SafeLoader)
     for key, value in defaults.items():
@@ -311,7 +408,8 @@ def request(ctx, filename=None, profile=None):
 
 def save_to_temp(tmp_dir, filename, content):
     filename = os.path.join(tmp_dir, filename)
-    print(f'saving {filename}')
+    log.info('saving %s', filename)
+    filename = filename.replace('*', 'star')
     if not os.path.exists(tmp_dir):
         os.makedirs(tmp_dir, 0o755, True)
     if isinstance(content, str):
@@ -346,7 +444,6 @@ def save_to_s3(ctx, bucket, ns, hostname, certificate, account_key, key,
     pfx_content = full_pfx(ctx, certificate, key)
     contents = [
         ('.crt', certificate),
-        ('.account_key', account_key),
         ('.key', key),
         ('.pfx', pfx_content),
     ]
@@ -354,8 +451,9 @@ def save_to_s3(ctx, bucket, ns, hostname, certificate, account_key, key,
     s3 = session.client('s3')
     for extension, content in contents:
         filename = f'{hostname}{extension}'
+        filename = filename.replace('*', 'star')
         filename = posixpath.join(ns, filename)
-        print(f'saving s3://{bucket}/{filename}')
+        log.info('saving s3://%s/%s', bucket, filename)
         if isinstance(content, str):
             content = content.encode('utf-8')
         s3.put_object(Bucket=bucket, Key=filename, Body=content)
@@ -376,7 +474,7 @@ def save_to_ssm(ctx, ns, hostname, certificate, account_key, key, profile=None):
     ]
     for suffix, content in contents:
         name = os.path.join(prefix, suffix)
-        print(f'saving {name}')
+        log.info('saving %s', name)
         save_chunked_ssm_parameter(ns, name, content, 'String', profile)
 
     contents = [
@@ -384,7 +482,7 @@ def save_to_ssm(ctx, ns, hostname, certificate, account_key, key, profile=None):
     ]
     for suffix, content in contents:
         name = os.path.join(prefix, suffix)
-        print(f'saving {name}')
+        log.info('saving %s', name)
         try:
             ssm.put_parameter(
                 Name=name,
@@ -394,7 +492,7 @@ def save_to_ssm(ctx, ns, hostname, certificate, account_key, key, profile=None):
                 Type='SecureString',
                 KeyId=f'alias/{ns}')
         except Exception as ex:  # pylint: disable=broad-except
-            print(f'exception saving to ssm: {ex}')
+            log.info('exception saving to ssm: %s', ex)
 
     name = os.path.join(prefix, 'pfx')
     save_chunked_ssm_parameter(ns, name, pfx_data, 'SecureString', profile)
@@ -409,7 +507,7 @@ def save_chunked_ssm_parameter(ns, name, value, type_, profile=None):
         value = value[4096:]
     for n, x in enumerate(pieces, 1):
         st = f'{name}{n}'
-        print(f'saving {st}')
+        log.info('saving %s', st)
         try:
             if type_ == 'SecureString':
                 ssm.put_parameter(
@@ -427,7 +525,7 @@ def save_chunked_ssm_parameter(ns, name, value, type_, profile=None):
                     Overwrite=True,
                     Type=type_)
         except Exception as ex:  # pylint: disable=broad-except
-            print(f'exception saving to ssm: {ex}')
+            log.info('exception saving to ssm: %s', ex)
 
 
 @task
@@ -520,3 +618,17 @@ def is_linux():
 
 def is_windows():
     return sys.platform == 'win32'
+
+
+@task
+def create_account_key(ctx, filename):
+    """
+    creates an account key and saves it to filename
+    """
+    import OpenSSL
+    key_type = OpenSSL.crypto.TYPE_RSA
+    key = OpenSSL.crypto.PKey()
+    key.generate_key(key_type, 2048)
+    st = OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_PEM, key)
+    with open(filename, 'w') as f:
+        f.write(st.decode())
